package data.scripts.world;

import com.fs.starfarer.api.Global;
import com.fs.starfarer.api.campaign.BaseCampaignEventListener;
import com.fs.starfarer.api.campaign.CampaignFleetAPI;
import com.fs.starfarer.api.campaign.LocationAPI;
import com.fs.starfarer.api.campaign.SectorAPI;
import com.fs.starfarer.api.campaign.SectorEntityToken;
import com.fs.starfarer.api.campaign.StarSystemAPI;
import com.fs.starfarer.api.campaign.econ.MarketAPI;
import com.fs.starfarer.api.fleet.FleetMemberAPI;
import com.fs.starfarer.api.impl.campaign.fleets.FleetFactory;
import com.fs.starfarer.api.util.Misc;
import com.fs.starfarer.api.util.WeightedRandomPicker;
import data.scripts.world.systems.SystemGenUtils;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import org.apache.log4j.Logger;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.lazywizard.lazylib.MathUtils;

public class AIW_VengeanceGeneratorManager extends BaseCampaignEventListener {

    protected static final String CONFIG_FILE = "data/config/AIWar/vengeanceGenConfig.json";
    
    public static final float RANDOM_MIN_ORBIT_RADIUS = 8000f;
    public static final float RANDOM_MAX_ORBIT_RADIUS = 15000f;
    public static final float RANDOM_MIN_HYPER_DISTANCE = 3000f;
    public static final float RANDOM_MAX_HYPER_DISTANCE = 12000f;
    
    public static int useRandomGenerators = 0;
    public static float randomSpawnChance = 0.5f;
    public static float randomSpawnChanceHyperspace = 1f;
    public static float baseSpawnMult = 1;
    public static float baseSpawnThreshold = 250f;
    public static float chargeEfficiency = 0.25f;
    public static float secondarySpawnMult = 0.5f;
    public static float secondaryChargeMult = 0.5f;
    
    protected static JSONArray generatorSystems;
      
    public static Logger log = Global.getLogger(AIW_VengeanceGeneratorManager.class);
    
    protected List<VengeanceGenerator> generators = new ArrayList<>();
    
    static {
        try {
            JSONObject config = Global.getSettings().loadJSON(CONFIG_FILE);
            useRandomGenerators = config.optInt("useRandomGenerators", useRandomGenerators);
            randomSpawnChance = (float)config.optDouble("randomSpawnChance", randomSpawnChance);
            randomSpawnChanceHyperspace = (float)config.optDouble("randomSpawnChanceHyperspace", randomSpawnChanceHyperspace);
            baseSpawnMult = (float)config.optDouble("baseSpawnMult", baseSpawnMult);
            baseSpawnThreshold = (float)config.optDouble("baseSpawnThreshold", baseSpawnThreshold);
            chargeEfficiency = (float)config.optDouble("chargeEfficiency", chargeEfficiency);
            secondarySpawnMult = (float)config.optDouble("secondarySpawnMult", secondarySpawnMult);
            secondaryChargeMult = (float)config.optDouble("secondaryChargeMult", secondaryChargeMult);
            generatorSystems = config.getJSONArray("systems");
        } catch (IOException | JSONException ex) {
            log.error(ex);
        }
    }
    
    
    public void spawnGenerator(LocationAPI system, SectorEntityToken toOrbit, float orbitRadius)
    {
        SectorEntityToken generator = system.addCustomEntity("aiw_vengeance_generator_" + (generators.size() + 1), null, "aiw_vengeance_generator", "darkspire");
        
        float orbitDays = SystemGenUtils.getOrbitalPeriod(toOrbit.getRadius(), orbitRadius, SystemGenUtils.getDensity(toOrbit));
        generator.setCircularOrbitWithSpin(toOrbit, MathUtils.getRandomNumberInRange(0, 360), orbitRadius, orbitDays, 30, 60);
        generators.add(new VengeanceGenerator(generator, system));
    }
    
    public void spawnHyperspaceGenerator(SectorAPI sector, float orbitRadius)
    {
        LocationAPI hyperspace = sector.getHyperspace();
        SectorEntityToken generator = hyperspace.addCustomEntity("aiw_vengeance_generator_" + (generators.size() + 1), null, "aiw_vengeance_generator", "darkspire");
        float orbitDays = SystemGenUtils.getOrbitalPeriod(500, orbitRadius, 1);
        generator.setCircularOrbitWithSpin(hyperspace.createToken(0, 0), MathUtils.getRandomNumberInRange(0, 360), orbitRadius, orbitDays, 30, 60);
        
        //float distance = MathUtils.getRandomNumberInRange(RANDOM_MIN_HYPER_DISTANCE, RANDOM_MAX_HYPER_DISTANCE);
        //float angle = MathUtils.getRandomNumberInRange(0f, 2*(float)Math.PI);
        //generator.setFixedLocation((float)Math.sin(angle)*distance, (float)Math.cos(angle)*distance);
        generators.add(new VengeanceGenerator(generator, hyperspace));
    }
    
    public AIW_VengeanceGeneratorManager() {
        super(true);
    }
    
    public void generate(SectorAPI sector, boolean wantRandom)
    {
        try {
            JSONObject config = Global.getSettings().loadJSON(CONFIG_FILE);
            int configRandom = config.optInt("useRandomGenerators", 0);
            
            boolean random = false;
            if (configRandom == -1) return;
            else if (configRandom == 0) random = wantRandom;
            else if (configRandom == 1) random = true;
            
            if (random)
            {
               List<StarSystemAPI> systems = sector.getStarSystems();
               for (StarSystemAPI system: systems)
               {
                   if (Math.random() > randomSpawnChance) continue;
                   SectorEntityToken toOrbit = system.getStar();
                   float orbitRadius = MathUtils.getRandomNumberInRange(RANDOM_MIN_ORBIT_RADIUS, RANDOM_MAX_ORBIT_RADIUS);
                   spawnGenerator(system, toOrbit, orbitRadius);
               }

               if (Math.random() <= randomSpawnChanceHyperspace)
               {
                   float orbitRadius = MathUtils.getRandomNumberInRange(RANDOM_MIN_HYPER_DISTANCE, RANDOM_MAX_HYPER_DISTANCE);
                   spawnHyperspaceGenerator(sector, orbitRadius);
               }
            }
            else
            {
                for(int i=0; i<generatorSystems.length(); i++)
                {
                    JSONObject entry = generatorSystems.getJSONObject(i);
                    String systemId = entry.getString("system");
                    
                    if (systemId.equals("hyperspace"))
                    {
                        float orbitRadius = (float)entry.optDouble("orbitRadius", 10000);
                        spawnHyperspaceGenerator(sector, orbitRadius);
                        continue;
                    }
                    
                    LocationAPI system = sector.getStarSystem(systemId);
                    if (system == null)
                    {
                        log.warn("Could not find system " + systemId);
                        continue;
                    }
                    String toOrbitId = entry.getString("orbits");
                    SectorEntityToken toOrbit = system.getEntityById(toOrbitId);
                    if (toOrbit == null)
                    {
                        log.warn("Could not find entity to orbit: " + toOrbitId);
                        continue;
                    }
                    
                    // make generator
                    float orbitRadius = (float)entry.optDouble("orbitRadius", 10000);
                    spawnGenerator(system, toOrbit, orbitRadius);
                }
            }
        } catch (IOException | JSONException ex) {
            log.error(ex);
        }
    }
    
    public void spawnVengeanceFleet(LocationAPI loc, SectorEntityToken spawner, float fp)
    {
        WeightedRandomPicker<SectorEntityToken> picker = new WeightedRandomPicker<>();
        if (!loc.isHyperspace())
        {
            List<MarketAPI> markets = Misc.getMarketsInLocation(loc);
            for (MarketAPI market : markets)
            {
                if (market.getFaction().isHostileTo("darkspire"))
                    picker.add(market.getPrimaryEntity(), market.getSize());
            }
        }
        if (picker.isEmpty())
        {
            List<MarketAPI> marketsAll = Global.getSector().getEconomy().getMarketsCopy();
            for (MarketAPI market : marketsAll)
            {
                if (market.getFaction().isHostileTo("darkspire"))
                    picker.add(market.getPrimaryEntity(), market.getSize());
            }
        }
        if (picker.isEmpty()) return;
        
        String name = "Vengeance Fleet";
        if (fp < 105) name = "Pestilence Fleet";
        else if (fp > 270) name = "Doom Fleet";
        
        CampaignFleetAPI fleet = FleetFactory.createGenericFleet("darkspire", name, 1, (int)fp);
        fleet.getMemoryWithoutUpdate().set("$fleetType", "spireVengeanceFleet");
        fleet.getMemoryWithoutUpdate().set("$maxFP", fp);
        loc.addEntity(fleet);
        
        fleet.setLocation(spawner.getLocation().x, spawner.getLocation().y);
        
        VengeanceFleetData data = new VengeanceFleetData(fleet);
        data.source = spawner;
        data.startingFleetPoints = fp;
        data.target = picker.pick();
        fleet.addScript(new AIW_VengeanceFleetAI(fleet, data));
    }

    @Override
    public void reportBattleOccurred(CampaignFleetAPI winner, CampaignFleetAPI loser) {
        float losses = 0f;
        
        if (winner.getFaction().getId().equals("darkspire") || loser.getFaction().getId().equals("darkspire"))
            return;
        
        boolean generatorPresent = false;
        for (VengeanceGenerator generator : generators)
        {
            if (generator.location == winner.getContainingLocation())
            {
                generatorPresent = true;
                break;
            }
        }
        if (!generatorPresent) return;
        
        List<FleetMemberAPI> loserCurrent = loser.getFleetData().getMembersListCopy();
        for (FleetMemberAPI member : loser.getFleetData().getSnapshot()) {
            if (!loserCurrent.contains(member)) {
                losses += member.getFleetPointCost();
            }
        }
        List<FleetMemberAPI> winnerCurrent = winner.getFleetData().getMembersListCopy();
        for (FleetMemberAPI member : winner.getFleetData().getSnapshot()) {
            if (!winnerCurrent.contains(member)) {
                losses += member.getFleetPointCost();
            }
        }
        losses *= chargeEfficiency;
        
        LocationAPI loc = winner.getContainingLocation();
        for (VengeanceGenerator generator : generators)
        {
            boolean isPrimary = (generator.location == loc);
            if (isPrimary) generator.charge += losses;
            else generator.charge += losses * (1/generators.size()) * secondaryChargeMult;
            if (generator.charge > baseSpawnThreshold)
            {
                float fpBase = generator.charge;
                for (VengeanceGenerator spawner: generators)
                {
                    float fp = fpBase;
                    if (spawner != generator) fp *= secondarySpawnMult;
                    fp *= MathUtils.getRandomNumberInRange(0.75f, 1.25f);
                    spawnVengeanceFleet(spawner.location, spawner.entity, fp);
                }
                generator.charge = 0;
            }
        }
    }
    
    public static class VengeanceGenerator
    {
        SectorEntityToken entity;
        float charge = 0;
        LocationAPI location;
        
        public VengeanceGenerator(SectorEntityToken entity, LocationAPI location)
        {
            this.entity = entity;
            this.location = location;
        }
    }
    
    public static class VengeanceFleetData
    {
        public CampaignFleetAPI fleet;
        public SectorEntityToken source;
        public SectorEntityToken target;
        public float startingFleetPoints = 0.0F;
    
        public VengeanceFleetData(CampaignFleetAPI fleet)
        {
            this.fleet = fleet;
        }
    }
}
